local HardFinger = _G.HardFinger
local P = HardFinger.HardFinger_Config

HardFinger.ENCOUNTERS_ID = {
["EMERALD"] = {1873, 1877, 1864}, -- EN
["TOV"] = {1958, 1962, 2008}, -- TOV
["NH"] = {1849, 1865, 1867, 1871, 1862, 1842, 1886, 1872, 1863, 1866}, -- NH
}

COMBATLOG_OBJECT_TYPE_PET		= COMBATLOG_OBJECT_TYPE_PET or 0x00001000
COMBATLOG_OBJECT_TYPE_GUARDIAN		= COMBATLOG_OBJECT_TYPE_GUARDIAN or 0x00002000
COMBATLOG_OBJECT_TYPE_NPC 		= COMBATLOG_OBJECT_TYPE_NPC or 0x0000800
PET_FILTER_BITMASK			= COMBATLOG_OBJECT_TYPE_PET + COMBATLOG_OBJECT_TYPE_GUARDIAN + COMBATLOG_OBJECT_TYPE_NPC
COMBATLOG_OBJECT_REACTION_FRIENDLY	= COMBATLOG_OBJECT_REACTION_FRIENDLY or 0x00000010

HardFinger.FailTable = {}
FailTable = HardFinger.FailTable
HardFinger.WhoPulledTable = {}
WhoPulledTable = HardFinger.WhoPulledTable
HardFinger.PetOwnersTable = {}
PetOwnersTable = HardFinger.PetOwnersTable
HardFinger.MobDeathsTable = {}
MobDeathsTable = HardFinger.MobDeathsTable

function HardFinger.TableHardCopy(orig)
	-- Credits to http://lua-users.org/wiki/CopyTable
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[HardFinger.TableHardCopy(orig_key)] = HardFinger.TableHardCopy(orig_value)
        end
        setmetatable(copy, HardFinger.TableHardCopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

function HardFinger.CombatToggle(status)
	if status == "OUT_OF_COMBAT" and UnitIsDeadOrGhost("player") == false then
		wipe(WhoPulledTable)

		wipe(PetOwnersTable)
		HardFinger.WorldInfo()
	elseif status == "IN_COMBAT" then
		--
	end

end

-- Funcao announce aura applied
function HardFinger.AnnounceAuraID(spellIDToTrack, message, destName, spellID)
	if message == "SPELL_AURA_APPLIED" or message == "SPELL_AURA_APPLIED_DOSE" then
		if spellIDToTrack == spellID then
			destName = HardFinger.SName(destName) -- Strip away realm name
			sendChat(destName.." is affected by "..GetSpellLink(spellID)..".", P.ReportChannel)
		end
	end
end

-- Funao AURA_APPLIED, AURA_APPLIED_DOSE

function badSpellAuraSpreadID(spellIDToTrack, safeSource, shouldShow, shouldTrack, message, spellID, sourceName, destName)
	if spellIDToTrack == spellID then
		if message == "SPELL_AURA_APPLIED" or message == "SPELL_AURA_APPLIED_DOSE" then
			sourceName = HardFinger.SName(sourceName) -- Strip away realm name
			destName = HardFinger.SName(destName) -- Strip away realm name
			if sourceName ~= safeSource then
				if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
					HardFinger.SendChat(sourceName.." failed miserably. "..GetSpellLink(spellID).." spreaded to "..destName..".", P.ReportChannel)
				end
				if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
					HardFinger.ModifyFailTable(spellID, "Bad Spread Count:", sourceName)
				end
			end
		end
	end
end

-- Has Interrupts
HardFinger.InterruptTable = {}
InterruptTable = HardFinger.InterruptTable
InterruptCDs = {[187707] = 15, -- Muzzle, Survival Hunter
 [1766] = 15, [57994] = 12, [31935] = 15, [183752] = 15, [147362] = 24, [96231] = 15,
    [15487] = 45, [6552] = 15, [32747] = 90}
function HardFinger.HasInterrupt(spellIDToTrack, message, spellID, sourceName, destName)
    if spellName == "Arcane Torrent" and message == "SPELL_CAST_SUCCESS" then
        InterruptTable[sourceName] = InterruptTable[sourceName] or {}
        InterruptTable[sourceName]["Arcane Torrent"] = GetTime()+90.5
    else
        for k,_ in pairs(InterruptCDs) do
            if k == spellID and message == "SPELL_CAST_SUCCESS" then
                InterruptTable[sourceName] = InterruptTable[sourceName] or {}
                InterruptTable[sourceName][spellID] = GetTime()+InterruptCDs[spellID]+0.5
                break
                
            end
        end
    end
    if spellIDToTrack == spellID and message == "SPELL_CAST_SUCCESS" then        
        HardFinger.DPrint(GetSpellLink(spellIDToTrack).. " cast finished.")
        local textPlayerNames = ""
        if InterruptTable ~= nil then
            for k,t in pairs(InterruptTable) do
                for spell,v in pairs(t) do
                    if GetTime() - v > 600 then
                        InterruptTable.k = nil
                    elseif GetTime() > v and UnitIsDeadOrGhost(k) == false then
                        HardFinger.ModifyFailTable(spellIDToTrack, "Missed Interrupts:", k)
                        if spell == "Arcane Torrent" then
                            textPlayerNames = textPlayerNames..HardFinger.SName(k).."(Racial), "
                        else
                            textPlayerNames = textPlayerNames..HardFinger.SName(k)..", "                                    
                        end
                    end
                end
            end
            if textPlayerNames ~= "" then
                textPlayerNames = textPlayerNames:sub(1,#textPlayerNames-2) -- Remove last ", "
                HardFinger.SendChat(textPlayerNames.." failed to interrupt "..GetSpellLink(spellID), P.ReportChannel)
            end
        end
    end
end


function HardFinger.WorldInfo()
	_,_,instanceType,_,_,_,_,_,instanceSize = GetInstanceInfo()
	if instanceType == 23 or instanceType == 8 then -- 5 man
		HardFinger.ProperTrigger = "DUNGEON"
	else
		HardFinger.ProperTrigger = "GENERIC"
	end
	if instanceType == 0 and not IsInGroup() then
		HardFinger.ProperChannel = "SAY"
	elseif instanceType == 17 then -- 17 = LFR
		HardFinger.ProperChannel = "INSTANCE_CHAT"
		HardFinger.ValidUnitFrameGUID = "raid"
	elseif instanceType >= 14 and instanceType <= 16 then -- 16 = Mythic Raid
		HardFinger.ProperChannel = "RAID"
		HardFinger.ValidUnitFrameGUID = "raid"
	else
		HardFinger.ProperChannel = "PARTY"
		HardFinger.ValidUnitFrameGUID = "party"
	end

	if instanceSize == 0 then
		HardFinger.InstanceSize = GetNumGroupMembers()-1
	end


	if HardFinger.ActiveEcounter ~= 0 then
		for k, v in pairs(HardFinger.ENCOUNTERS_ID) do
			if HardFinger.IsIn(HardFinger.ActiveEncounter, v) then
				HardFinger.ProperTrigger = k
				break
			end
		end
	end
end

function HardFinger.IsIn(object, table)
    if type(table) == "string" then
        return object == table
    end
    local result = false
    for _,v in pairs(table) do
        if v == object then
            result = true
            break
        end
        
    end
    return result
end

--message, _, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellID, spellName, param12, param13, param14 = ...

function HardFinger.BadSpellHitID(spellToTrack, amountThreshold, roles, shouldShow, shouldTrack, message, destName, destFlags, spellID, damageAmount, overKillAmount, absorbAmount)
	if spellToTrack == spellID and message == "SPELL_DAMAGE" and bit.band(destFlags, PET_FILTER_BITMASK) == 0 then
		local role = UnitGroupRolesAssigned(destName)
		if string.find(string.upper(roles),string.upper(role)) or roles == "ALL" then
			local absorbed = absorbAmount or 0
			local overkill = overKillAmount or 0
			local damageTaken = damageAmount + absorbed
		
			if damageTaken >= amountThreshold then
				if shouldShowDamage == true then
					HardFinger.SendChat(destName.." failed miserably. "..GetSpellLink(spellID).." hit for "..HardFinger.FormatDamage(damageTaken)..".", P.ReportChannel)
				end
				HardFinger.ModifyFailTable(spellID, "Total times hit: ", destName)
			end
		end
	end
end

function HardFinger.BadSpellHitIDDeath(spellToTrack, amountThreshold, roles, shouldShow, shouldTrack, message, destName, destFlags, spellID, damageAmount, overKillAmount, absorbAmount)
	if spellToTrack == spellID and message == "SPELL_DAMAGE" and bit.band(destFlags, PET_FILTER_BITMASK) == 0 then
		local overkill = overKillAmount or 0
		if overkill > 0 then
			destName = HardFinger.SName(destName) -- Strip away realm name
			local role = UnitGroupRolesAssigned(destName)
			if string.find(string.upper(roles),string.upper(role)) or roles == "ALL" then
				local absorbed = absorbAmount or 0
				local damageTaken = damageAmount + absorbed
	
				if damageTaken >= amountThreshold then
					if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
						HardFinger.SendChat(destName.." failed miserably. "..GetSpellLink(spellID).." Overkill: "..HardFinger.FormatDamage(damageTaken)..".", P.ReportChannel)
					end
					if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
						HardFinger.ModifyFailTable(spellID, "Deaths: ", destName)
					end
				end
			end
		end
	end
end

function HardFinger.toggletf(variable)
	if variable ~= true then		return true	else		return false	end
end

function HardFinger.toggle01(variable)
	if variable ~= 1 then		return 1	else		return 0	end
end


function HardFinger.StatusText(variable)
	if variable ~= 1 and variable ~= true then
		return "|cffff0000Off"
	else
		return "|c0000FF00On"
	end
end

-- Funcao remove from table

function HardFinger.RemoveFromTableIDbyEvent(tableToModify, event, spellIDToTrack, message, tableKey, spellID)

	if tableToModify == nil then
		return

	end

	if spellIDToTrack == spellID and string.find(event,message) then
		if tableKey == nil then

			HardFinger.DPrint("key = "..tableKey.." does not exists.")
		return

	end

	tableToModify[tableKey] = nil

end
end

function HardFinger.FindOwnerPetFromGUID(petName,petGUID) -- Credits to Recount

	local ownerName
	local ownerGUID
	local ownerRealm
	if PetOwnersTable[petGUID] ~= nil then
		return PetOwnersTable[petGUID][1], PetOwnersTable[petGUID][2]
    end

	if petGUID==UnitGUID("pet") then
		ownerName = UnitName("player")

		ownerGUID = UnitGUID("player")
		PetOwnersTable[petGUID] = {ownerName, ownerGUID}
		return ownerName, ownerGUID

	end

	if HardFinger.InstanceSize > 0 then
		for i = 1, HardFinger.InstanceSize, 1 do -- GetNumRaidMembers()
			if petGUID == UnitGUID("raidpet"..i) then
				ownerName, ownerRealm = UnitName("raid"..i)

				if ownerRealm then
					ownerName = ownerName .. "-" .. ownerRealm

				end
				ownerGUID = UnitGUID("raid"..i)
				PetOwnersTable[petGUID] = {ownerName, ownerGUID}
				return ownerName, ownerGUID

			end

		end

	end

	HardFinger.DPrint(petName.." owner not found")
	return petName, petGUID

end

function HardFinger.DamageToTarget(targetNameToTrack, desc, message, sourceGUID, sourceName, sourceFlags, destName, spellID, damageAmount, overKillAmount, absorbAmount)
	if destName == targetNameToTrack and (message == "SPELL_DAMAGE" or message == "SPELL_PERIODIC_DAMAGE" or message == "SWING_DAMAGE") then
		local absorbed, damaged, damageDone
		if message == "SPELL_DAMAGE" or message == "SPELL_PERIODIC_DAMAGE" then
			absorbed = absorbAmount or 0
			damaged = damageAmount or 0
		else
			absorbed = absorbAmount or 0
			damaged = spellID or 0 -- For melee swings, damageAmount is the same nth parameter as spellID
			spellName = "Melee Swing"
		end
		damageDone = damaged + absorbed

		local sourceDamage = sourceName
		if bit.band(sourceFlags, PET_FILTER_BITMASK) ~= 0 then
			sourceDamage,_ = HardFinger.FindOwnerPetFromGUID(sourceName, sourceGUID)
		end
	
		if desc ~= nil and sourceDamage ~= nil and damageDone ~= nil then -- Not necessary, but just to be safe
			HardFinger.ModifyFailTable(desc, "Damage Done ["..targetNameToTrack.."]: ", sourceDamage, damageDone)
		end
	end
end

-- Funao AURA_APPLIED, AURA_APPLIED_DOSE, param14 = amount

function HardFinger.BadSpellAuraID(spellIDToTrack, stackAmount, modulo, shouldShow, shouldTrack, message, destName, spellID, numberOfStacks)
	if spellIDToTrack == spellID then
		if message == "SPELL_AURA_APPLIED" or message == "SPELL_AURA_APPLIED_DOSE" then
			destName = HardFinger.SName(destName) -- Strip away realm name
			if stackAmount == -1 then
				if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
					HardFinger.SendChat(destName.." has failed miserably. "..GetSpellLink(spellID))
				end
				if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
					HardFinger.ModifyFailTable(spellID, "Total # applications: ", destName, 1)
				end
			elseif numberOfStacks ~= nil and numberOfStacks > stackAmount then
				if modulo == nil or modulo == -1 then
					if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
						HardFinger.SendChat(destName.." has failed miserably. "..GetSpellLink(spellID).." "..numberOfStacks.." stacks.", pChannel)
					end
					if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
						HardFinger.ModifyFailTable(spellID, "Highest # stacks: ", destName, nil, numberOfStacks)
					end
				elseif stackAmount >= numberOfStacks and mod(numberOfStacks-stackAmount, modulo) == 0 then
					if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
						HardFinger.SendChat(destName.." has failed miserably. "..GetSpellLink(spellID).." "..numberOfStacks.." stacks.", pChannel)
					end
					if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
						HardFinger.ModifyFailTable(spellID, "Highest # stacks: ", destName, nil, numberOfStacks)
					end

				end

			end
		end

	end

end


function HardFinger.DisplayFailTable(call)
	--HardFinger.DPrint("Entered function displayFailTable")
	if call == "HRESET" then
		wipe(FailTable or {})
		HardFinger.DPrint("Failtable HReset")
		return
	end
	-- Check whether FailTable is empty, if so, return
	local next = next
	if next(FailTable) == nil then
		HardFinger.DPrint("failTable is empty")
		HardFinger.FailTableDelayTimer = nil
		return
	else
		HardFinger.FailTableDelayTimer = HardFinger.FailTableDelayTimer or GetTime()
		if (GetTime() - HardFinger.FailTableDelayTimer) > 1 then
			HardFinger.DPrint("failTable being displayed")
			for spell,spellTable in pairs(FailTable) do
				local counter = ""
				if HardFinger.IsIn(spellTable.description, {"Total times hit: ", "Deaths: "}) == true then
					counter, counter2 = HardFinger.SumVectorValues(spellTable.players)
				end
				local counter2text = ""
				if counter2 ~= nil then
					-- counter2text = "#Players: "..counter2
				end
				-- spell Icon
				spellIconTextOrLink = spellTable.spellIcon
				if type(spellTable.spellIcon) == "number" then
					spellIconTextOrLink = (GetSpellLink(spellTable.spellIcon))
				end
				HardFinger.SendChat(spellIconTextOrLink.." "..(spellTable.description)..counter..counter2text, P.ReportChannel)
				local textRows = {[1] = ""}
				local index = 1
				for _,unit in HardFinger.Spairs(spellTable.players, function(t,a,b) return t[b].counter < t[a].counter end) do
					if #textRows[index] > 220 then
						index = index + 1
					end
					textToAdd = HardFinger.SName(unit.name).."  "..HardFinger.FormatDamage(unit.counter).. ", "
					textRows[index] = (textRows[index] or "")..textToAdd
				end
				textRows[index] = textRows[index]:sub(1,#textRows[index]-2) -- Remove last ", " from the last textRow
				for i=1, #textRows do

					HardFinger.SendChat(textRows[i], P.ReportChannel)
				end
			end
		HardFinger.FailTableDelayTimer = nil
		end
	end
end

-- Funao bad SPELL_DAMAGE spread
function HardFinger.BadSpreadDamageID(spellToTrack, unitName, amountThreshold, roles, shouldShow, shouldTrack, message, sourceName, destName, spellID, damageTaken, overkill)
    if spellToTrack == spellID then
        if message == "SPELL_DAMAGE" or message == "SPELL_ABSORB" then
            absorbed = absorbed or 0
            damageTaken = damageTaken + absorbed
            sourceName = HardFinger.SName(sourceName) -- Strip away realm name
            destName = HardFinger.SName(destName) -- Strip away realm name
            if sourceName ~= destName then
                if sourceName ~= unitName then
                    local role = UnitGroupRolesAssigned(sourceName)
                    if string.find(string.upper(roles),string.upper(role)) or roles == "ALL" then
                        if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
                            HardFinger.SendChat(sourceName.." hit "..destName.." "..GetSpellLink(spellID).." "..damageTaken, P.ReportChannel)                
                        end   
                        if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
                            HardFinger.ModifyFailTable(spellID, "Spread Fail [Friendly Fire]: ", sourceName, damageTaken)
                        end
                    end
                end
            end
        end
    end
end


function HardFinger.BadSpellDamageID(spellIDToTrack, amountThreshold, roles, shouldShow, shouldTrack, message, destName, spellID, damageAmount, overKillAmount, absorbAmount)
	if spellIDToTrack == spellID and UnitIsPlayer(destName) == true then
		if message == "SPELL_DAMAGE" or message == "SPELL_PERIODIC_DAMAGE" then
			local role = UnitGroupRolesAssigned(destName)
			destName = HardFinger.SName(destName) -- Strip away realm name
			if string.find(string.upper(roles),string.upper(role)) or roles == "ALL" then
				local absorbed = absorbAmount or 0
				local overkill = overKillAmount or 0
				local damageTaken = damageAmount + absorbed
				if damageTaken >= amountThreshold then
					if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
						HardFinger.SendChat(destName.." failed miserably. "..GetSpellLink(spellID).." hit for "..HardFinger.FormatDamage(damageTaken)..".", P.ReportChannel)
					end
					if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
						HardFinger.ModifyFailTable(spellID, "Total damage taken: ", destName, damageTaken)
					end
				end
			end
		end
	end
end

function HardFinger.Spairs(t, order)
	-- collect the keys
	local keys = {}
	for k in pairs(t) do
		keys[#keys+1] = k
	end
	-- if order function given, sort by it by passing the table and keys a, b,
	-- otherwise just sort the keys
	if order then
		table.sort(keys, function(a,b) return order(t, a, b) end)
	else
		table.sort(keys)
	end
	-- return the iterator function
	local i = 0
	return function()
		i = i + 1
		if keys[i] then
			return keys[i], t[keys[i]]
		end
	end
end

--Funcao Announce Dispel

function HardFinger.AnnounceDispel(spellToTrack, message, sourceName, destName, debuff)
	if message == "SPELL_DISPEL" then
		if param14 == spellToTrack then
			sourceName = HardFinger.SName(sourceName) -- Strip away realm name
			destName = HardFinger.SName(destName) -- Strip away realm name
			HardFinger.SendChat(sourceName.." has dispeled "..debuff.." from "..destName..".", P.ReportChannel)
		end

	end

end


function HardFinger.MobDeaths(Reset, SendReport, mobName, message, spellID, destName, param14)
	if Reset == true then
		MobDeathsTable[mobName] = 0
	elseif SendReport == true and MobDeathsTable[mobName] ~= nil then
		HardFinger.SendChat("Total "..mobName.." killed: "..MobDeathsTable[mobName]..".", P.ReportChannel)
		wipe(MobDeathsTable)
	elseif message == "SPELL_DAMAGE" then
		if destName == mobName and param14 > 0 and MobDeathsTable[mobName] ~= nil then
			MobDeathsTable[mobName] = MobDeathsTable[mobName] + 1
		end
	end
end

function HardFinger.TauntRezLust(message, sourceName, sourceRaidFlags, destName, destRaidFlags, spellID)
	-- 6795 = Growl, 62124 = Hand of Reckoning, 355 = Taunt, 116189 = Provoke, 185245 = Torment, 56222 = Dark Command
	if P.Taunt == true and HardFinger.IsIn(spellID, {6795, 62124, 355, 116189, 185245, 56222}) then
		-- 6795 = Growl, 62124 = Hand of Reckoning, 355 = Taunt, 116189 = Provoke, 185245 = Torment, 56222 = Dark Command
        if message == "SPELL_AURA_APPLIED" then
			rIcon = HardFinger.GetRaidMarkByFlags(destRaidFlags)
			destName = HardFinger.SName(destName)
			sourceName = HardFinger.SName(sourceName)
			HardFinger.SendChat(sourceName.." "..GetSpellLink(spellID).."'d "..rIcon..destName..rIcon..".", P.ReportChannel)
		end
	elseif message == "SPELL_CAST_SUCCESS" then
		--61999 = Raise Ally, 20484 = Rebirth, 20707 = Soulstone, 21169 = Reincarnation
		if P.BattleRez == true and HardFinger.IsIn(spellID, {61999, 20484, 20707, 21169}) then
			if spellID == 21169 then
				rIcon = HardFinger.GetRaidMarkByFlags(sourceRaidFlags)
				HardFinger.SendChat(rIcon..sourceName..rIcon.." used "..GetSpellLink(spellID)..".", P.ReportChannel)
			else
				rIcon = HardFinger.GetRaidMarkByFlags(destRaidFlags)
				HardFinger.SendChat(sourceName.." used "..GetSpellLink(spellID).." on "..rIcon..destName..rIcon..".", P.ReportChannel)
			end
		--2825 = Blood Lust, 80353 = Time Warp, 178207 = Drums of Fury, 230935 = Drums of the Mountain
		elseif P.BloodLust and HardFinger.IsIn(spellID, {2825, 80353, 178207, 230935}) then
			HardFinger.SendChat(sourceName.." used "..GetSpellLink(spellID)..".", P.ReportChannel)
		end
	end
end

function HardFinger.WhoRezzed(message, sourceName, destName, destRaidFlags, spellID)
	if message == "SPELL_CAST_SUCCESS" then
		--61999 = Raise Ally, 20484 = Rebirth, 20707 = Soulstone
        	if HardFinger.IsIn(spellID, {61999, 20484, 20707}) then
				rIcon = HardFinger.GetRaidMarkByFlags(destRaidFlags)
				HardFinger.SendChat(sourceName.." used "..GetSpellLink(spellID).." on "..rIcon..destName..rIcon..".", P.ReportChannel)
			end
	end
end

function HardFinger.WhoLusted(message, sourceName, spellID)
	if message == "SPELL_CAST_SUCCESS" then
		--2825 = Blood Lust, 80353 = Time Warp, 178207 = Drums of Fury
        	if HardFinger.IsIn(spellID, {2825, 80353, 178207}) then
					HardFinger.SendChat(sourceName.." used "..GetSpellLink(spellID)..".", P.ReportChannel)
		end
	end
end

function HardFinger.WhoPulled(name, beforeTimer, message, sourceGUID, sourceName, destGUID, destName, destFlags, spellID)
	if WhoPulledTable[destGUID] == nil then
		if (HardFinger.IsIn(destName, name) or name == "ALL") and spellID ~= 137619 and spellID ~= 203528  then
			if message == "SPELL_DAMAGE" or message == "SPELL_AURA_APPLIED" or message == "SPELL_AURA_APPLIED_DOSE" then
				if string.find(sourceGUID, "Player") or string.find(sourceGUID, "Pet") then
					if bit.band(destFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY) == 0 then
						if sourceName ~= destName then
							WhoPulledTable[destGUID] = true
							-- Get the name of who pulled this mob
							local wpname = sourceName

							-- If its a pet, get owner instead
							if bit.band(sourceFlags, PET_FILTER_BITMASK) ~= 0 then
								wpname = wpname.." ("..HardFinger.SName(HardFinger.FindOwnerPetFromGUID(sourceName, sourceGUID)).."'s pet)"
							end
	                            
							local text = HardFinger.SName(wpname).." aggroed "..destName.." with "..GetSpellLink(spellID)
							if beforeTimer == true and GetTime() < (HardFinger.DBMPullTimer or 0)-0.1 then
								local diff = HardFinger.DBMPullTimer - GetTime()
								diff = format("%.1fs", diff)
								text = text.." ("..diff.." before DBM Pull)"
							end
							if name == "ALL" then
								HardFinger.SendChat(text, P.WhoPulledChannel)
							else
								HardFinger.SendChat(text, P.ReportChannel)
							end
						end
					end
				end
			end
		end
	end
end

-- Funo sName (short name, strip away realm name)
function HardFinger.SName(name)
	if name:find("-") == nil then
		return name
	else
		return name:match("(.-)-")
	end
end


function HardFinger.BuffDebuffAfterEvent(spellIDToTrack, messageToTrack, roles, func, auraName, spellIDIcon, textDesc, shouldShow, shouldTrack, message, spellID)
    if spellID == spellIDToTrack and message == messageToTrack then
		local _,_,_,_,_,_,_,_,instanceSize = GetInstanceInfo() -- Needed to handle if people leave the group
		HardFinger.InstanceSize = instanceSize
		for i = 1, HardFinger.InstanceSize, 1 do
			local iName = UnitName(HardFinger.ValidUnitFrameGUID..i)
			local role = UnitGroupRolesAssigned(iName)
			if string.find(string.upper(roles),string.upper(role)) or roles == "ALL" then
				if UnitIsDeadOrGhost(iName) == false then
					if func(iName, auraName) == nil then
						if shouldShow == true or shouldShow == 2 or shouldShow == 3 then
							HardFinger.SendChat(iName.." failed miserably and got hit by "..GetSpellLink(spellID)..".", P.ReportChannel)
						end
						if shouldTrack == true or shouldTrack == 1 or shouldTrack == 3 then
							HardFinger.ModifyFailTable(spellIDIcon, textDesc, iName)
						end
					end
				end
			end
        end
    end
end

function HardFinger.HealAbsorbed(message, hasBuff, sourceName, destName, healabsorb)

	healabsorb = healabsorb or 0

	if message == "SPELL_HEAL" then

		if UnitDebuff(destName, hasBuff) ~= nil then

			if UnitGroupRolesAssigned(sourceName) == "HEALER" then

				--HardFinger.DPrint(sourceName, "Healed", destName, "for", healabsorb, param13, param14, param15)
				HardFinger.ModifyFailTable(206610, "Healing Absorbed: ", HardFinger.SName(sourceName), healabsorb)

			end
		end
	end

end


-- Modificar fail Table, descricao = texto no chat

function HardFinger.ModifyFailTable(spellIconOrText, description, creatureName, amountToChange, highestAmount)
	local amountToChange = amountToChange or highestAmount or 1
	local spelln = spellIconOrText
	if type(spelln) == "number" then
		spelln = GetSpellInfo(spellIconOrText)
	end
	local uniqueTableKey = spelln..description

	--HardFinger.DPrint("received: "..uniqueTableKey, spellIcon, description, creatureName, amountToChange)
	FailTable[uniqueTableKey] = FailTable[uniqueTableKey] or {}
	FailTable[uniqueTableKey].description = description
	FailTable[uniqueTableKey].spellIcon = spellIconOrText
	FailTable[uniqueTableKey].players = FailTable[uniqueTableKey].players or {}
	FailTable[uniqueTableKey].players[creatureName] = FailTable[uniqueTableKey].players[creatureName] or {}
	FailTable[uniqueTableKey].players[creatureName].name = creatureName
	FailTable[uniqueTableKey].players[creatureName].counter = FailTable[uniqueTableKey].players[creatureName].counter or 0
	FailTable[uniqueTableKey].players[creatureName].counter = FailTable[uniqueTableKey].players[creatureName].counter + amountToChange
	
	if highestAmount ~= nil then
		if highestAmount > FailTable[uniqueTableKey].players[creatureName].counter then
			FailTable[uniqueTableKey].players[creatureName].counter = highestAmount
		end
	end
	amountToChange = nil

end



-- Funao Send Chat
function HardFinger.SendChat(string, channel)
    if channel == false then
		print(string)
	elseif channel == "GUILD_CHANNEL" then
		SendChatMessage(string, "GUILD")
	elseif (tonumber(P.ReportChannel) or 0) ~= 1 then
		SendChatMessage(string, "CHANNEL", nil, P.ReportChannel)
	else
		SendChatMessage(string, HardFinger.ProperChannel)
	end
end

-- Somar valores para vetor
function HardFinger.SumVectorValues(vector)

	local sum = 0

	local totalEntries = 0

	for _, v in pairs(vector) do
		sum = sum + v.counter

		totalEntries = totalEntries + 1

	end
	return sum, totalEntries

end

-- Formatar damage properly
function HardFinger.FormatDamage(damage)
    local ret = damage
    if damage > 1e6 then
        ret = format("%.1f M", damage / 1e6) -- million
    elseif damage > 1e3 then
        ret = format("%.1f K", damage / 1e3) -- thousand
    end
    return ret
end

HardFinger.SterilizeTable = {}
SterilizeTable = HardFinger.SterilizeTable
function HardFinger.SterilizeUptime(message, sourceName, destName, spellID)
	if HardFinger.LastEncounter == 1867 and HardFinger.ActiveEncounter == 0 then
		forceEnd = true
	end

	if message == "SPELL_AURA_APPLIED" then -- or message == "SPELL_AURA_REFRESH" then
		if spellID == 208499 or spellID == 211615 then
			SterilizeTable[destName] = SterilizeTable[destName] or {}
			SterilizeTable[destName].count = SterilizeTable[destName].count or 0
			SterilizeTable[destName].buffUptime = SterilizeTable[destName].buffUptime or 0
			SterilizeTable[destName].lastApplied = GetTime()
		end

	end

	if SterilizeTable ~= nil and SterilizeTable ~= {} then
		if spellID == 212103 and (message == "SPELL_DAMAGE" or message == "SPELL_PERIODIC_DAMAGE" or message == "SPELL_ABSORB") then
			SterilizeTable[destName].count = SterilizeTable[destName].count + 1
		end
		if (message == "SPELL_AURA_REMOVED" and (spellID == 208499 or spellID == 211615)) then
			balance = GetTime() - SterilizeTable[destName].lastApplied
			SterilizeTable[destName].buffUptime = SterilizeTable[destName].buffUptime + balance
		end
		if forceEnd == true then
			-- for player, innerTable in pairs(sterilizeTable) do
			--     balance = GetTime() - innerTable.lastApplied
			--     innerTable.buffUptime = innerTable.buffUptime + balance
			-- end
			forceEnd = false
			HardFinger.ComputeSterilize()
		end
	end
end

function HardFinger.DPrint(args)
	if UnitName("player") == "Eumesminha" or UnitName("player") == "Camoleze" then
		print(args)
	end
end

function HardFinger.ComputeSterilize()
	for player, innerTable in pairs(HardFinger.SterilizeTable) do
		HardFinger.DPrint("Computing for "..player)
		HardFinger.DPrint("\nComputing for player..", player)
		result = innerTable.count * 2.5 / innerTable.buffUptime * 100
		result = format("%.1f", result)
		HardFinger.ModifyFailTable(208499, "Score! (higher is better): ", HardFinger.SName(player), nil, tonumber(result))
	end
	wipe(SterilizeTable)
end



-- Funo raid target flags ICONs
function HardFinger.GetRaidMarkByFlags(flags)
    if flags == 0 then
        return "" 
    end
    local r = log(flags)/log(2) + 1
    markName = "NONE"
    if r == 1 then
        markName = "STAR"
    elseif r == 6 then
        markName = "SQUARE"
    elseif r == 7 then
        markName = "CROSS"
    elseif r == 8 then
        markName = "SKULL"
    end
    return "{rt"..r.."}",  MarkName
end

-- Funo HP Porcentagem
function HardFinger.Hpp(unit)
	return math.max(0, UnitHealth(unit)) / math.max(1, UnitHealthMax(unit)) * 100;
end